volatility_filter =f"""
%s
r_vol = rank(ts_std_dev(returns, 10));
enter_cond = if_else(r_vol > 0.5, 1, -1);
trade_when(enter_cond, %s, -1)
"""

def reduce_turnover(expr: str, final_factor: str):
    """
    Reduce turnover by adding volatility filter
    Effective for price-volume based factors
    :param expr: origin factor expression
    :param final_factor: name of the final factor
    :return:
    """
    return volatility_filter % (expr, final_factor)


if __name__ == "__main__":
    expr0 = "a = rank(((-1 * returns) * ts_mean(volume, 20) * (high - low)));"
    print(reduce_turnover(expr0, "a"))


import dataclasses

from brainminer.utils import is_empty_metrics
from typing import Any


@dataclasses.dataclass
class AlphaAggregateData:

    def __init__(self, alpha_id: str):
        self.alpha_id = alpha_id
        self.drawdown = None
        self.turnover = None
        self.sharpe = None
        self.margin = None
        self.fitness = None
        self.returns = None

    def extract_metrics(self, metrics):
        self.sharpe = metrics['sharpe']
        self.turnover = metrics['turnover']
        self.drawdown = metrics['drawdown']
        self.margin = metrics['margin']
        self.fitness = metrics['fitness']
        self.returns = metrics['returns']

class PerformanceFilter:

    def __init__(self, fitness = 1.0, sharpe = 1.25, turnover = -1.0, drawdown = -1.0, returns = 0.2):
        self._fitness = fitness
        self._sharpe = sharpe
        self._turnover = turnover
        self._drawdown = drawdown
        self._returns = returns

    @property
    def fitness(self):
        return self._fitness

    @property
    def sharpe(self):
        return self._sharpe

    @property
    def turnover(self):
        return self._turnover

    @property
    def drawdown(self):
        return self._drawdown

    @property
    def returns(self):
        return self._returns

    def is_match(self, metrics) -> bool:
        if self._fitness > 0.0 and metrics['fitness'] < self._fitness:
            return False
        if self._sharpe > 0.0 and metrics['sharpe'] < self._sharpe:
            return False
        if 0.0 < self._turnover < metrics['turnover']:
            return False
        if 0.0 < self._drawdown < metrics['drawdown']:
            return False
        if self._returns > 0.0 and metrics['returns'] < self._returns:
            return False
        return True


class AlphaSimulationResult:

    def __init__(self, success: bool, message: str, expr: str, sim_id: str, sim_data: Any, alpha_data: Any):
        self.success = success
        self.message = message
        self.expression = expr
        self.sim_id = sim_id
        self.sim_data = sim_data
        self.alpha_data = alpha_data

    @staticmethod
    def fail(expr: str, sim_id: str, message: str):
        return AlphaSimulationResult(success=False, sim_id=sim_id, message=message, expr=expr, sim_data=None,
                                     alpha_data=None)

    @staticmethod
    def success(expr: str, sim_id: str, sim_data: Any, alpha_data: Any):
        return AlphaSimulationResult(success=True, sim_id=sim_id, message='', expr=expr, sim_data=sim_data,
                                     alpha_data=alpha_data)

    def get_alpha_id(self):
        return '' if self.alpha_data is None else self.alpha_data['id']

    def is_match(self, perf_filter: PerformanceFilter):
        if self.success and self.alpha_data is not None:
            m = perf_filter.is_match(self.alpha_data['is'])
            if not m:
                return False
            if not is_empty_metrics(self.alpha_data['os']):
                return perf_filter.is_match(self.alpha_data['os'])
        return False




from typing import List, Dict

from brainminer.brain.model import AlphaAggregateData, AlphaSimulationResult, PerformanceFilter


class Evaluator:

    @staticmethod
    def get_fitness_level(fitness: float, delay: int):
        if delay == 0:
            if fitness > 3.25:
                return 'Spectacular'
            elif fitness > 2.6:
                return 'Excellent'
            elif fitness > 1.95:
                return 'Good'
            elif fitness > 1.3:
                return 'Average'
            else:
                return 'Needs Improvement'
        else:
            if fitness > 2.5:
                return 'Spectacular'
            elif fitness > 2:
                return 'Excellent'
            elif fitness > 1.5:
                return 'Good'
            elif fitness > 1:
                return 'Average'
            else:
                return 'Needs Improvement'

    @staticmethod
    def get_is_metrics(alpha_data: Dict):
        return alpha_data['is']

    @staticmethod
    def extract_aggregate_data(alpha_data: Dict) -> AlphaAggregateData:
        metrics = alpha_data['is']
        data = AlphaAggregateData(alpha_id=alpha_data['id'])
        data.extract_metrics(metrics)
        return data

    @staticmethod
    def load_is_fail_checks(alpha_data: Dict) -> List[str]:
        is_metrics = alpha_data['is']
        fail_checks = []
        for item in is_metrics['checks']:
            if item['result'] == 'FAIL':
                reason = "{}:{} is below cutoff of {}".format(item['name'], item["value"], item["limit"])
                fail_checks.append(reason)
        return fail_checks

    @staticmethod
    def filter_alphas(alpha_result: List[AlphaSimulationResult], perf_filter: PerformanceFilter) -> List[str]:
        match_alphas = []
        for item in alpha_result:
            if item.success and item.alpha_data is not None and perf_filter.is_match(item.alpha_data['is']):
                match_alphas.append(item.alpha_data['id'])
        return match_alphas






import json
import logging
import os
import time

from brainminer.brain.session import SessionManager
from brainminer.brain.model import PerformanceFilter
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Dict

# Configure logger
logger = logging.getLogger(__name__)

class AlphaManager:

    def __init__(self, session_manager: SessionManager):
        self._session_manager = session_manager

    # def filter_alphas(self, perf_filter: PerformanceFilter) -> List[str]:
    #     match_alphas = []
    #     for item in alpha_result:
    #         if item.success and item.alpha_data is not None and perf_filter.is_match(item.alpha_data['is']):
    #             match_alphas.append(item.alpha_data['id'])
    #     return match_alphas

    def fet_successful_alphas(self, offset: int = 0, limit: int = 10):
        """Fetch successful unsubmitted alphas with good performance metrics."""
        url = "https://api.worldquantbrain.com/users/self/alphas"
        params = {
            "limit": limit,
            "offset": offset,
            "status": "UNSUBMITTED",
            "is.fitness>": 1,
            "is.sharpe>": 1.25,
            "order": "-dateCreated",
            "hidden": "false"
        }
        logger.info(f"Fetching alphas with params: {params}")
        full_url = f"{url}?{'&'.join(f'{k}={v}' for k,v in params.items())}"
        logger.info(f"Request URL: {full_url}")

        max_retries = 3
        retry_delay = 60
        for attempt in range(max_retries):
            try:
                logger.debug(f"Attempt {attempt + 1}/{max_retries} to fetch alphas")
                response = self._session_manager.get(url, params=params)
                logger.info(f"Response URL: {response.url}")
                logger.info(f"Response status: {response.status_code}")
                logger.info(f"Response headers: {dict(response.headers)}")
                logger.info(f"Response content: {response.text[:1000]}...")  # First 1000 chars

                if response.status_code == 429:  # Too Many Requests
                    wait_time = int(response.headers.get('Retry-After', retry_delay))
                    logger.info(f"Rate limited. Waiting {wait_time} seconds...")
                    time.sleep(wait_time)
                    continue

                response.raise_for_status()
                data = response.json()
                logger.info(
                    f"Successfully fetched {len(data.get('results', []))} alphas. Total count: {data.get('count', 0)}")
                return data

            except Exception as e:
                if attempt < max_retries - 1:
                    logger.warning(f"Attempt {attempt + 1} failed: {str(e)}")
                    logger.warning(f"Response URL: {response.url if 'response' in locals() else 'N/A'}")
                    logger.warning(f"Response text: {response.text if 'response' in locals() else 'N/A'}")
                    time.sleep(retry_delay)
                else:
                    logger.error(f"Failed to fetch alphas after {max_retries} attempts. Last error: {e}")
                    logger.error(f"Last response URL: {response.url if 'response' in locals() else 'N/A'}")
                    logger.error(f"Last response text: {response.text if 'response' in locals() else 'N/A'}")
                    return {"count": 0, "results": []}

        return {"count": 0, "results": []}

    def monitor_submission(self, alpha_id: str, max_attempts: int = 30, sleep_time: int = 10) -> Dict:
        """Monitor submission status until complete or failed."""
        url = f"https://api.worldquantbrain.com/alphas/{alpha_id}/submit"

        for attempt in range(max_attempts):
            try:
                response = self.sess.get(url)
                logger.info(f"Monitoring attempt {attempt + 1} for alpha {alpha_id}")
                logger.info(f"Response status: {response.status_code}")
                logger.info(f"Response content: {response.text[:1000]}...")

                if response.status_code != 200:
                    logger.error(f"Submission likely failed for alpha {alpha_id}")
                    logger.error(f"Response status: {response.status_code}")
                    logger.error(f"Response text: {response.text}")
                    return {"status": "failed", "error": response.text}

                # If response is empty (still submitting)
                if not response.text.strip():
                    logger.info(f"Alpha {alpha_id} still being submitted, waiting...")
                    time.sleep(sleep_time)
                    continue

                # Try to parse JSON response (submission complete)
                try:
                    data = response.json()
                    logger.info(f"Submission complete for alpha {alpha_id}")
                    return data
                except json.JSONDecodeError:
                    logger.info(f"Response not in JSON format yet for alpha {alpha_id}, continuing to monitor...")

            except Exception as e:
                logger.warning(f"Monitor attempt {attempt + 1} failed: {str(e)}")
                logger.warning(f"Response content: {response.text if 'response' in locals() else 'N/A'}")

            time.sleep(sleep_time)

        logger.error(f"Monitoring timed out for alpha {alpha_id}")
        return {"status": "timeout", "error": "Monitoring timed out"}

    def log_submission_result(self, alpha_id: str, result: Dict) -> None:
        """Log submission result to file."""
        log_file = 'submission_results.json'

        # Load existing results
        existing_results = []
        if os.path.exists(log_file):
            try:
                with open(log_file, 'r') as f:
                    existing_results = json.load(f)
            except json.JSONDecodeError:
                logger.warning(f"Could not parse {log_file}, starting fresh")

        # Add new result
        entry = {
            "alpha_id": alpha_id,
            "timestamp": int(time.time()),
            "result": result
        }
        existing_results.append(entry)

        # Save updated results
        with open(log_file, 'w') as f:
            json.dump(existing_results, f, indent=2)

        logger.info(f"Logged submission result for alpha {alpha_id}")

    def has_fail_checks(self, alpha: Dict) -> bool:
        """Check if alpha has any FAIL results in checks."""
        checks = alpha.get("is", {}).get("checks", [])
        return any(check.get("result") == "FAIL" for check in checks)

    def submit_alpha(self, alpha_id: str) -> bool:
        """Submit a single alpha and monitor its status."""
        url = f"https://api.worldquantbrain.com/alphas/{alpha_id}/submit"
        logger.info(f"Submitting alpha {alpha_id}")
        logger.info(f"Request URL: {url}")

        try:
            # Initial submission
            response = self.sess.post(url)
            logger.info(f"Response status: {response.status_code}")

            if response.status_code == 201:
                logger.info(f"Successfully submitted alpha {alpha_id}, monitoring status...")

                # Monitor submission status
                result = self.monitor_submission(alpha_id)
                if result:
                    self.log_submission_result(alpha_id, result)
                    return True
                else:
                    logger.error(f"Submission monitoring timed out for alpha {alpha_id}")
                    return False
            else:
                logger.error(f"Failed to submit alpha {alpha_id}. Status: {response.status_code}")
                logger.error(f"Response text: {response.text}")
                return False

        except Exception as e:
            logger.error(f"Error submitting alpha {alpha_id}: {str(e)}")
            logger.exception("Full traceback:")
            return False

    def batch_submit(self, batch_size: int = 5) -> None:
        """Submit alphas in batches with pagination."""
        logger.info(f"Starting batch submission with batch size {batch_size}")
        offset = 0
        total_submitted = 0

        while True:
            logger.info(f"Fetching batch at offset {offset}")
            response = self.fetch_successful_alphas(offset=offset, limit=batch_size)

            if not response or not response.get("results"):
                logger.info("No more alphas to process")
                break

            results = response["results"]
            if not results:
                logger.info("Empty results batch")
                break

            logger.info(f"Processing batch of {len(results)} alphas...")

            # Filter out alphas with FAIL checks
            valid_alphas = [alpha for alpha in results if not self.has_fail_checks(alpha)]
            logger.info(f"Found {len(valid_alphas)} valid alphas after filtering FAILs")

            if not valid_alphas:
                logger.info("No valid alphas in this batch, moving to next")
                offset += batch_size
                continue

            # Submit valid alphas in parallel
            with ThreadPoolExecutor(max_workers=3) as executor:
                futures = []
                for alpha in valid_alphas:
                    alpha_id = alpha["id"]
                    expression = alpha["regular"]["code"]
                    metrics = (f"Sharpe: {alpha['is']['sharpe']}, "
                               f"Fitness: {alpha['is']['fitness']}")
                    logger.info(f"Queuing alpha {alpha_id} for submission:")
                    logger.info(f"Expression: {expression}")
                    logger.info(f"Metrics: {metrics}")
                    futures.append(executor.submit(self.submit_alpha, alpha_id))

                for future in as_completed(futures):
                    if future.result():
                        total_submitted += 1

            if not response.get("next"):
                logger.info("No more pages to process")
                break

            offset += batch_size
            logger.info(f"Waiting 60 seconds before next batch...")
            time.sleep(60)

        logger.info(f"Submission process complete. Total alphas submitted: {total_submitted}")



import logging
import pandas as pd

from brainminer.brain import Simulator, SessionManager, AlphaManager, SimulatorSettings, AlphaSimulationResult
from brainminer.utils import RateLimiter
from typing import List, Dict

logger = logging.getLogger(__name__)

class WorldQuantBrain:

    def __init__(self, session_manager: SessionManager = None):
        if session_manager is None:
            session_manager = SessionManager()
        self._session_manager = session_manager
        self._simulator = Simulator(session_manager=session_manager)
        self._alpha_submitter = AlphaManager(session_manager=session_manager)

    def execute_simulation(self, expr: str, settings: SimulatorSettings = None) -> AlphaSimulationResult:
        return self._simulator.execute(expression=expr, settings=settings)

    def execute_multi_simulation(self, expression_list: List[str], settings: SimulatorSettings = None, rate_limiter: RateLimiter = None) -> List[AlphaSimulationResult]:
        return self._simulator.execute_multi(expression_list=expression_list, settings=settings, rate_limiter=rate_limiter)

    def load_alpha_result(self, alpha_id: str):
        return self._simulator.load_alpha_result(alpha_id)

    def get_operators(self) -> List[Dict]:
        """Fetch available operators from WorldQuant Brain."""
        try:
            print("Requesting operators...")
            response = self._session_manager.get('https://api.worldquantbrain.com/operators')

            response.raise_for_status()
            data = response.json()

            # Handle different response formats
            if isinstance(data, list):
                return data
            return data.get('results', [])
        except Exception as e:
            logger.error(f"Failed to fetch operators: {e}")
            return []

    def set_alpha_properties(self,
                             alpha_id,
                             name: str = None,
                             color: str = None,
                             selection_desc: str = "None",
                             combo_desc: str = "None",
                             tags: List = None,
                             ):
        """
        Function changes alpha's description parameters
        """

        if tags is None:
            tags = []
        params = {
            "color": color,
            "name": name,
            "tags": tags,
            "category": None,
            "regular": {"description": None},
            "combo": {"description": combo_desc},
            "selection": {"description": selection_desc},
        }
        response = self._session_manager.patch(
            "https://api.worldquantbrain.com/alphas/" + alpha_id, json=params
        )

    def get_datafields(self,instrument_type: str = 'EQUITY',
                       region: str = 'USA', delay: int = 1,
                       universe: str = 'TOP3000', dataset_id: str = '', search: str = '') -> pd.DataFrame:
        """
        get available datafields
        data columns: id | description | dataset | category | subcategory | region | delay | universe | type | coverage | userCount | alphaCount | themes |
        """
        if len(search) == 0:
            url_template = "https://api.worldquantbrain.com/data-fields?" +\
                f"&instrumentType={instrument_type}" +\
                f"&region={region}&delay={str(delay)}&universe={universe}&dataset.id={dataset_id}&limit=50" +\
                "&offset={x}"
            count = self._session_manager.get(url_template.format(x=0)).json()['count']
        else:
            url_template = "https://api.worldquantbrain.com/data-fields?" +\
                f"&instrumentType={instrument_type}" +\
                f"&region={region}&delay={str(delay)}&universe={universe}&limit=50" +\
                f"&search={search}" +\
                "&offset={x}"
            count = 100

        datafields_list = []
        for x in range(0, count, 50):
            url = url_template.format(x=x)
            datafields = self._session_manager.get(url)
            datafields_list.append(datafields.json()['results'])

        datafields_list_flat = [item for sublist in datafields_list for item in sublist]

        datafields_df = pd.DataFrame(datafields_list_flat)
        return datafields_df

# if __name__ == '__main__':
#     brain = WorldQuantBrain()
#     res = brain.load_alpha_result(alpha_id='1NMg3dz')
#     print(res)


import datetime
import threading
import time

class RateLimiter:
    def __init__(self, max_requests: int = 10, time_window: int = 60):
        self.max_requests = max_requests
        self.time_window = time_window  # in seconds
        self.requests = []
        self.lock = threading.Lock()

    def can_make_request(self) -> bool:
        now = datetime.datetime.now()
        with self.lock:
            # Remove old requests
            self.requests = [t for t in self.requests
                             if t > now - datetime.timedelta(seconds=self.time_window)]

            if len(self.requests) < self.max_requests:
                self.requests.append(now)
                return True
            return False

    def wait_for_slot(self):
        while not self.can_make_request():
            time.sleep(1)


# from typing import List

# TODO 没有ts_min, ts_max操作符
ALPHA_101_FORMULAS = [
    # Original formulas adapted for WorldQuant Brain syntax
    "rank(ts_arg_max(power(if_else(returns < 0, ts_std_dev(returns, 20), close), 2), 5)) - 0.5",
    "-1 * ts_corr(rank(ts_delta(log(volume), 2)), rank((close - open) / open), 6)",
    "-1 * ts_corr(rank(open), rank(volume), 10)",
    "-1 * ts_rank(rank(low), 9)",
    "rank(open - (ts_mean(vwap, 10))) * (-1 * abs(rank(close - vwap)))",
    "-1 * ts_corr(open, volume, 10)",
    "if_else(ts_mean(volume, 20) < volume, -1 * ts_rank(abs(ts_delta(close, 7)), 60) * sign(ts_delta(close, 7)), -1)",
    "-1 * rank((ts_sum(open, 5) * ts_sum(returns, 5)) - ts_delay((ts_sum(open, 5) * ts_sum(returns, 5)), 10))",
    "if_else(0 < ts_min(ts_delta(close, 1), 5), ts_delta(close, 1), if_else(ts_max(ts_delta(close, 1), 5) < 0, ts_delta(close, 1), -1 * ts_delta(close, 1)))",
    "rank(if_else(0 < ts_min(ts_delta(close, 1), 4), ts_delta(close, 1), if_else(ts_max(ts_delta(close, 1), 4) < 0, ts_delta(close, 1), -1 * ts_delta(close, 1))))",
    "(rank(ts_max(vwap - close, 3)) + rank(ts_min(vwap - close, 3))) * rank(ts_delta(volume, 3))",
    "sign(ts_delta(volume, 1)) * (-1 * ts_delta(close, 1))",
    "-1 * rank(ts_covariance(rank(close), rank(volume), 5))",
    "-1 * rank(ts_delta(returns, 3)) * ts_corr(open, volume, 10)",
    "-1 * ts_sum(rank(ts_corr(rank(high), rank(volume), 3)), 3)",
    "-1 * rank(ts_covariance(rank(high), rank(volume), 5))",
    "-1 * rank(ts_rank(close, 10)) * rank(ts_delta(ts_delta(close, 1), 1)) * rank(ts_rank(volume / ts_mean(volume, 20), 5))",
    "-1 * rank(ts_std_dev(abs(close - open), 5) + (close - open) + ts_corr(close, open, 10))",
    "-1 * sign(ts_delta(close, 7) + ts_delta(close, 7)) * (1 + rank(1 + ts_sum(returns, 250)))",
    "-1 * rank(open - ts_delay(high, 1)) * rank(open - ts_delay(close, 1)) * rank(open - ts_delay(low, 1))",

    # Next batch (21-40)
    "if_else((ts_mean(close, 8) / 8 + ts_std_dev(close, 8)) < (ts_mean(close, 2) / 2), -1, if_else((ts_mean(close, 2) / 2) < ((ts_mean(close, 8) / 8) - ts_std_dev(close, 8)), 1, if_else(1 < (volume / ts_mean(volume, 20)), 1, -1)))",
    "-1 * (ts_delta(ts_corr(high, volume, 5), 5) * rank(ts_std_dev(close, 20)))",
    "if_else((ts_mean(high, 20) / 20) < high, -1 * ts_delta(high, 2), 0)",
    "if_else((ts_delta(ts_mean(close, 100) / 100, 100) / ts_delay(close, 100)) < 0.05, -1 * (close - ts_min(close, 100)), -1 * ts_delta(close, 3))",
    "rank(((-1 * returns) * ts_mean(volume, 20) * vwap * (high - close)))",
    "-1 * ts_max(ts_corr(ts_rank(volume, 5), ts_rank(high, 5), 5), 3)",
    "if_else(0.5 < rank((ts_sum(ts_corr(rank(volume), rank(vwap), 6), 2) / 2.0)), -1, 1)",
    "scale(((ts_corr(ts_mean(volume, 20), low, 5) + ((high + low) / 2)) - close))",
    "(min(product(rank(rank(scale(log(ts_sum(ts_min(rank(rank(-1 * rank(ts_delta(close - 1, 5)))), 2), 1)))), 1), 5) + ts_rank(ts_delay(-1 * returns, 6), 5))",
    "(1.0 - rank(((sign(close - ts_delay(close, 1)) + sign(ts_delay(close, 1) - ts_delay(close, 2))) + sign(ts_delay(close, 2) - ts_delay(close, 3))))) * ts_sum(volume, 5) / ts_sum(volume, 20)",
    "(rank(rank(rank(ts_decay_linear(-1 * rank(rank(ts_delta(close, 10))), 10)))) + rank(-1 * ts_delta(close, 3))) + sign(scale(ts_corr(ts_mean(volume, 20), low, 12)))",
    "scale((ts_mean(close, 7) / 7 - close)) + (20 * scale(ts_corr(vwap, ts_delay(close, 5), 230)))",
    "rank(-1 * (1 - (open / close)))",
    "rank((1 - rank(ts_std_dev(returns, 2) / ts_std_dev(returns, 5))) + (1 - rank(ts_delta(close, 1))))",
    "(ts_rank(volume, 32) * (1 - ts_rank((close + high - low), 16))) * (1 - ts_rank(returns, 32))",
    "(((2.21 * rank(ts_corr((close - open), ts_delay(volume, 1), 15))) + (0.7 * rank((open- close)))) + (0.73 * rank(ts_rank(ts_delay(-1 * returns, 6), 5)))) + rank(abs(ts_corr(vwap, ts_mean(volume, 20), 6))) + (0.6 * rank((ts_mean(close, 200) / 200 - open) * (close - open)))",
    "rank(ts_corr(ts_delay(open - close, 1), close, 200)) + rank(open - close)",
    "-1 * rank(ts_rank(close, 10)) * rank(close / open)",
    "-1 * rank(ts_delta(close, 7) * (1 - rank(ts_decay_linear(volume / ts_mean(volume, 20), 9))))) * (1 + rank(ts_sum(returns, 250)))",
    "-1 * rank(ts_std_dev(high, 10)) * ts_corr(high, volume, 10)",

    # Next batch (41-60)
    "power((high * low), 0.5) - vwap",
    "rank((vwap - close)) / rank((vwap + close))",
    "ts_rank((volume / ts_mean(volume, 20)), 20) * ts_rank((-1 * ts_delta(close, 7)), 8)",
    "-1 * ts_corr(high, rank(volume), 5)",
    "-1 * ((rank(ts_mean(ts_delay(close, 5), 20) / 20) * ts_corr(close, volume, 2)) * rank(ts_corr(ts_sum(close, 5), ts_sum(close, 20), 2)))",
    "if_else(0.25 < (ts_delta(ts_delay(close, 20), 10) / 10 - ts_delta(ts_delay(close, 10), 10) / 10), -1, if_else((ts_delta(ts_delay(close, 20), 10) / 10 - ts_delta(ts_delay(close, 10), 10) / 10) < 0, 1, -1 * ts_delta(close, 1)))",
    "((rank(1 / close) * volume / ts_mean(volume, 20)) * ((high * rank(high - close)) / (ts_mean(high, 5) / 5))) - rank(ts_delta(vwap, 5))",
    "(ts_corr(ts_delta(close, 1), ts_delta(ts_delay(close, 1), 1), 250) * ts_delta(close, 1)) / close",
    "if_else((ts_delta(ts_delay(close, 20), 10) / 10 - ts_delta(ts_delay(close, 10), 10) / 10) < -0.1, 1, -1 * ts_delta(close, 1))",
    "-1 * ts_max(rank(ts_corr(rank(volume), rank(vwap), 5)), 5)",

    # Batch (61-80)
    "if_else((ts_delta(ts_delay(close, 20), 10) / 10 - ts_delta(ts_delay(close, 10), 10) / 10) < -0.05, 1, -1 * ts_delta(close, 1))",
    "-1 * ((ts_min(low, 5) - ts_delay(ts_min(low, 5), 5)) * rank((ts_sum(returns, 240) - ts_sum(returns, 20)) / 220)) * ts_rank(volume, 5)",
    "-1 * ts_delta((((close - low) - (high - close)) / (close - low)), 9)",
    "-1 * ((low - close) * power(open, 5)) / ((low - high) * power(close, 5))",
    "-1 * ts_corr(rank((close - ts_min(low, 12)) / (ts_max(high, 12) - ts_min(low, 12))), rank(volume), 6)",
    "-1 * (rank((ts_sum(returns, 10) / ts_sum(ts_sum(returns, 2), 3))) * rank(returns))",
    "-1 * ((close - vwap) / ts_decay_linear(rank(ts_arg_max(close, 30)), 2))",
    "-1 * ts_rank(ts_decay_linear(ts_corr(vwap, volume, 4), 8), 6)",
    "-1 * ts_rank(ts_decay_linear(ts_corr(((vwap * 0.728317) + (vwap * (1 - 0.728317))), volume, 4), 16), 8)",
    "-1 * (2 * scale(rank(((close - low) - (high - close)) / (high - low) * volume)) - scale(rank(ts_arg_max(close, 10))))",
    "if_else(rank(vwap - ts_min(vwap, 16)) < rank(ts_corr(vwap, ts_mean(volume, 180), 18)), 1, -1)",
    "if_else(rank(ts_corr(vwap, ts_sum(ts_mean(volume, 20), 22), 10)) < rank(((rank(open) + rank(open)) < (rank((high + low) / 2) + rank(high)))), 1, -1)",
    "-1 * (rank(ts_decay_linear(ts_delta(close, 2), 8)) - rank(ts_decay_linear(ts_corr(((vwap * 0.318108) + (open * (1 - 0.318108))), ts_sum(ts_mean(volume, 180), 37), 14), 12)))",
    "-1 * if_else(rank(ts_corr(ts_sum(((open * 0.178404) + (low * (1 - 0.178404))), 13), ts_sum(ts_mean(volume, 120), 13), 17)) < rank(ts_delta(((high + low) / 2 * 0.178404) + (vwap * (1 - 0.178404)), 4)), 1, -1)",
    "-1 * if_else(rank(ts_corr(((open * 0.00817205) + (vwap * (1 - 0.00817205))), ts_sum(ts_mean(volume, 60), 9), 6)) < rank(open - ts_min(open, 14)), 1, -1)",
    "-1 * (rank(ts_decay_linear(ts_delta(vwap, 4), 7)) + ts_rank(ts_decay_linear(((low * 0.96633) + (low * (1 - 0.96633)) - vwap) / (open - (high + low) / 2), 11), 7))",
    "-1 * power(rank(high - ts_min(high, 2)), rank(ts_corr(vwap, ts_mean(volume, 120), 6)))",
    "-1 * if_else(ts_rank(ts_corr(rank(high), rank(ts_mean(volume, 15)), 9), 14) < rank(ts_delta((close * 0.518371) + (low * (1 - 0.518371)), 1)), 1, -1)",
    "-1 * power(rank(ts_max(ts_delta(vwap, 3), 5)), ts_rank(ts_corr(((close * 0.490655) + (vwap * (1 - 0.490655))), ts_mean(volume, 20), 5), 9))",
    "-1 * power(rank(ts_delta(vwap, 1)), ts_rank(ts_corr(close, ts_mean(volume, 50), 18), 18))",

    # Batch (81-101)
    "max(ts_rank(ts_decay_linear(ts_corr(ts_rank(close, 3), ts_rank(ts_mean(volume, 180), 12), 18), 4), ts_rank(ts_decay_linear(rank((low + open) - (vwap + vwap)), 16), 4))",
    "rank(ts_decay_linear(ts_corr((high + low) / 2, ts_mean(volume, 40), 9), 10)) / rank(ts_decay_linear(ts_corr(ts_rank(vwap, 4), ts_rank(volume, 19), 7), 3))",
    "max(rank(ts_decay_linear(ts_delta(vwap, 5), 3)), ts_rank(ts_decay_linear(((ts_delta(((open * 0.147155) + (low * (1 - 0.147155))), 2) / ((open * 0.147155) + (low * (1 - 0.147155)))) * -1), 3), 17)) * -1",
    "-1 * if_else(rank(ts_corr(close, ts_sum(ts_mean(volume, 30), 37), 15)) < rank(ts_corr(rank((high * 0.0261661) + (vwap * (1 - 0.0261661))), rank(volume), 11)), 1, -1)",
    "if_else(rank(ts_corr(vwap, volume, 4)) < rank(ts_corr(rank(low), rank(ts_mean(volume, 50)), 12)), 1, -1)",
    "-1 * max(rank(ts_decay_linear(ts_delta(vwap, 1), 12)), ts_rank(ts_decay_linear(ts_rank(ts_corr(low, ts_mean(volume, 81), 8), 20), 17), 19))",
    "min(rank(ts_decay_linear(((high + low) / 2 + high - (vwap + high)), 20)), rank(ts_decay_linear(ts_corr((high + low) / 2, ts_mean(volume, 40), 3), 6)))",
    "power(rank(ts_corr(ts_sum(((low * 0.352233) + (vwap * (1 - 0.352233))), 20), ts_sum(ts_mean(volume, 40), 20), 7)), rank(ts_corr(rank(vwap), rank(volume), 6)))",
    "if_else(rank(ts_delta(((close * 0.60733) + (open * (1 - 0.60733))), 1)) < rank(ts_corr(ts_rank(vwap, 4), ts_rank(ts_mean(volume, 150), 9), 15)), 1, -1)",
    "-1 * power(rank(sign(ts_delta(((open * 0.868128) + (high * (1 - 0.868128))), 4))), ts_rank(ts_corr(high, ts_mean(volume, 10), 5), 6))",
    "-1 * if_else(rank(log(product(rank(ts_corr(vwap, ts_sum(ts_mean(volume, 10), 50), 8)), 15))) < rank(ts_corr(rank(vwap), rank(volume), 5)), 1, -1)",
    "-1 * min(rank(ts_decay_linear(ts_delta(open, 1), 15)), ts_rank(ts_decay_linear(ts_corr(volume, ((open * 0.634196) + (open * (1 - 0.634196))), 17), 7), 13))",
    "((rank(ts_delay((high - low) / (ts_sum(close, 5) / 5), 2)) * rank(rank(volume))) / ((high - low) / (ts_sum(close, 5) / 5)) / (vwap - close))",
    "power(ts_rank((vwap - ts_max(vwap, 15)), 21), ts_delta(close, 5))",
    "power(rank(ts_corr(((high * 0.876703) + (close * (1 - 0.876703))), ts_mean(volume, 30), 10)), rank(ts_corr(ts_rank((high + low) / 2, 4), ts_rank(volume, 7), 7)))",
    "-1 * if_else(ts_rank(ts_corr(close, ts_sum(ts_mean(volume, 20), 15), 6), 20) < rank((open + close) - (vwap + open)), 1, -1)",
    "-1 * max(rank(ts_decay_linear(ts_delta(((close * 0.369701) + (vwap * (1 - 0.369701))), 2), 3)), ts_rank(ts_decay_linear(abs(ts_corr(ts_mean(volume, 81), close, 13)), 5), 14))",
    "min(rank(ts_decay_linear(((rank(open) + rank(low)) - (rank(high) + rank(close))), 8)), ts_rank(ts_decay_linear(ts_corr(ts_rank(close, 8), ts_rank(ts_mean(volume, 60), 21), 8), 7), 3))",
    "(ts_rank(ts_decay_linear(ts_corr(((low * 0.967285) + (low * (1 - 0.967285))), ts_mean(volume, 10), 7), 6), 4) - ts_rank(ts_decay_linear(ts_delta(vwap, 3), 10), 15))",
    "-1 * power(rank(close - ts_max(close, 5)), ts_rank(ts_corr(ts_mean(volume, 40), low, 5), 3))",
    "-1 * (ts_rank(ts_decay_linear(ts_decay_linear(ts_corr(close, volume, 10), 16), 4), 5) - rank(ts_decay_linear(ts_corr(vwap, ts_mean(volume, 30), 4), 3)))",
    "min(ts_rank(ts_decay_linear(((high + low) / 2 + close < (low + open)), 15), 19), ts_rank(ts_decay_linear(ts_corr(rank(low), rank(ts_mean(volume, 30)), 8), 7), 7))",
    "(ts_rank(ts_decay_linear(ts_corr(vwap, ts_mean(volume, 81), 17), 20), 8) / rank(ts_decay_linear(ts_delta(((close * 0.524434) + (vwap * (1 - 0.524434))), 3), 16)))",
    "-1 * power(rank(vwap - ts_min(vwap, 12)), ts_rank(ts_corr(ts_rank(vwap, 20), ts_rank(ts_mean(volume, 60), 4), 18), 3))",
    "if_else(rank(open - ts_min(open, 12)) < ts_rank(rank(ts_corr(ts_sum((high + low) / 2, 19), ts_sum(ts_mean(volume, 40), 19), 13)), 12), 1, -1)",
    "-1 * max(ts_rank(ts_decay_linear(ts_corr(rank(vwap), rank(volume), 4), 4), 8), ts_rank(ts_decay_linear(ts_arg_max(ts_corr(ts_rank(close, 7), ts_rank(ts_mean(volume, 60), 4), 13), 14), 13))",
    "-1 * (rank(ts_decay_linear(ts_delta(((low * 0.721001) + (vwap * (1 - 0.721001))), 3), 20)) - ts_rank(ts_decay_linear(ts_rank(ts_corr(ts_rank(low, 8), ts_rank(ts_mean(volume, 60), 17), 5), 19), 7))",
    "(rank(ts_decay_linear(ts_corr(vwap, ts_sum(ts_mean(volume, 5), 26), 5), 7)) - rank(ts_decay_linear(ts_rank(ts_arg_min(ts_corr(rank(open), rank(ts_mean(volume, 15)), 21), 9), 8)))",
    "-1 * if_else(rank(ts_corr(ts_sum((high + low) / 2, 20), ts_sum(ts_mean(volume, 60), 20), 9)) < rank(ts_corr(low, volume, 6)), 1, -1)",
    "-1 * (1.5 * scale(rank(((close - low) - (high - close)) / (high - low) * volume)) - scale(rank(ts_arg_min(close, 30))))",
    "divide(subtract(close, open), add(subtract(high, low), 0.001))"
]


import dataclasses
import logging
import os
import time
import threading
from typing import Optional

import requests
import yaml
from requests.auth import HTTPBasicAuth


@dataclasses.dataclass
class UserCredentials(object):
    user_id: str
    username: str
    password: str

logging.basicConfig(encoding='utf-8', level=logging.INFO, format='%(asctime)s: %(message)s')
logger = logging.getLogger(__name__)

class SessionManager:

    def __init__(self, config_file = None):
        self._sessions = dict()
        self._user: Optional[UserCredentials] = None
        self._session: Optional[requests.Session] = None
        self._last_login_time = 0
        self._lock = threading.Lock()
        self.load_config(config_file)

    def load_config(self, config_file):
        if config_file is None or config_file == '':
            config_file = os.path.expanduser('~/.brain/config.yaml')
        assert os.path.exists(config_file)
        with open(config_file, 'r') as f:
            data = yaml.safe_load(f)
            user_data = data.get('user', {})
            self._user = UserCredentials(**user_data)

    def execute_login(self):
        try:
            self._lock.acquire()
            if self._session is not None:
                return self._session
            cur_thread = threading.current_thread().name
            logger.info(f"Execute login in thread {cur_thread}")
            session = requests.Session()
            session.auth = HTTPBasicAuth(self._user.username, self._user.password)
            resp = session.post('https://api.worldquantbrain.com/authentication')
            if resp.status_code != 201:
                raise Exception(f"Authentication failed: {resp.text}")
            json_resp = resp.json()
            if 'user' in json_resp and 'id' in json_resp['user']:
                assert json_resp['user']['id'] == self._user.user_id
                logger.info("Successfully authenticated with WorldQuant Brain")
                self._last_login_time = time.time()
                self._session = session
            return self._session
        except Exception as e:
            logger.error(f"Execute authentication failed: {str(e)}")
            raise e
        finally:
            self._lock.release()


    def get_session(self) -> Optional[requests.Session]:
        if self._session is None:
            self.execute_login()
        return self._session


    def get(self, url, params=None, headers=None):
        sess = self.get_session()
        resp = sess.get(url, params=params, headers=headers)
        if resp.status_code == 401:
            sess = self.execute_login()
            return sess.get(url, params=params, headers=headers)
        else:
            return resp

    def post(self, url, data=None, json=None, **kwargs):
        sess = self.get_session()
        resp = sess.post(url, data=data, json=json, **kwargs)
        if resp.status_code == 401:
            sess = self.execute_login()
            return sess.post(url, data=data, json=json, **kwargs)
        else:
            return resp

    def patch(self, url, data=None, **kwargs):
        sess = self.get_session()
        resp = sess.patch(url, data=data, **kwargs)
        if resp.status_code == 401:
            sess = self.execute_login()
            return sess.patch(url, data=data, **kwargs)
        else:
            return resp


if __name__ == '__main__':
    m = SessionManager()
    session = m.get_session()



from .session import SessionManager
from .simulator import Simulator, SimulatorSettings, Universe, Neutralization
from .alpha import AlphaManager
from .evaluator import Evaluator
from .model import AlphaSimulationResult, PerformanceFilter
from .brain import WorldQuantBrain


import json
import logging
import time
from concurrent.futures.thread import ThreadPoolExecutor

from brainminer.brain import SessionManager
from enum import Enum
from typing import List, Optional

from brainminer.brain.model import AlphaSimulationResult
from brainminer.utils import RateLimiter

logger = logging.getLogger(__name__)

class SimulationSubmitResult(object):

    def __init__(self, success: bool, message: str, sim_id: str):
        self.success = success
        self.message = message
        self.sim_id = sim_id

    @staticmethod
    def fail(message: str):
        return SimulationSubmitResult(message=message, success=False, sim_id='')

    @staticmethod
    def success(sim_id: str):
        return SimulationSubmitResult(sim_id=sim_id, success=True, message='Success')

class Universe(Enum):
    TOP3000 = 1
    TOP1000 = 2
    TOP500 = 3
    TOP200 = 4
    TOPSP500 = 5

class Neutralization(Enum):
    NONE = 1
    MARKET = 2
    SECTOR = 3
    INDUSTRY = 4
    SUBINDUSTRY = 5

class SimulatorSettings:

    def __init__(self):
        self._config = {
            'instrumentType': 'EQUITY',
            'region': 'USA',
            'universe': 'TOP3000',
            'delay': 1,
            'decay': 0,
            'neutralization': 'INDUSTRY',
            'truncation': 0.08,
            'pasteurization': 'ON',
            'unitHandling': 'VERIFY',
            'nanHandling': 'OFF',
            'language': 'FASTEXPR',
            'visualization': False,
        }

    @property
    def config(self):
        return self._config

    def set_delay(self, delay: int):
        self._config['delay'] = delay

    def set_decay(self, decay: int):
        self._config['decay'] = decay

    def set_universe(self, uni: Universe):
        self._config['universe'] = uni.name

    def set_neutralization(self, neut: Neutralization):
        self._config['neutralization'] = neut.name

    def update(self, **kwargs):
        self._config.update(kwargs)

class Simulator:

    def __init__(self, session_manager: SessionManager):
        self._session_manager = session_manager

    def submit_simulation(self, expression: str, settings: SimulatorSettings = None) -> SimulationSubmitResult:
        """
        submit simulation
        :param expression: fast expression script
        :param settings: simulation settings
        :return: simulation submit result
        """
        if settings is None:
            settings = SimulatorSettings()
        simulation_data = {
            'type': 'REGULAR',
            'settings': settings.config,
            'regular': expression
        }

        while True:
            try:
                logger.info(f"Submitting alpha with expression: {expression}")
                response = self._session_manager.post("https://api.worldquantbrain.com/simulations", json=simulation_data)
                logger.info(f"Response status: {response.status_code}")

                if response.status_code == 201:
                    sim_id = response.headers.get('Location', '').split('/')[-1]
                    if sim_id:
                        logger.info(f"Simulation created with ID: {sim_id}")
                        return SimulationSubmitResult.success(sim_id)
                    else:
                        return SimulationSubmitResult.fail("No simulation ID in response headers")
                elif response.status_code == 400:
                    error_data = response.json()
                    logger.error(f"API error details: {error_data}")
                    error_msg = error_data.get("error", {}).get("message", "Unknown error")
                    return SimulationSubmitResult.fail(error_msg)
                elif response.status_code == 429:
                    retry_after_sec = float(response.headers.get("Retry-After", 0))
                    wait_time = max(retry_after_sec, 10.0)
                    logger.info(f"Rate limited.Waiting {wait_time} seconds...")
                    time.sleep(wait_time)
                else:
                    response.raise_for_status()
                    return SimulationSubmitResult.fail("Unknown error, response = " + response.text)
            except Exception as e:
                logger.error(f"Error testing alpha {expression}: {str(e)}")
                return SimulationSubmitResult.fail(str(e))

    def monitor_simulation(self, sim_id: Optional[str], expression: str) -> AlphaSimulationResult:
        """Monitor simulation progress and get final results."""
        if sim_id is None or sim_id == '':
            return AlphaSimulationResult.fail(expr=expression, sim_id='', message="Empty simulation id")
        try:
            sim_data = self.wait_for_simulation_complete(sim_id)
            # If completed, check alpha details
            if sim_data.get("status") in ("COMPLETE", "WARNING") and sim_data.get("alpha"):
                alpha_id = sim_data["alpha"]
            else:
                logger.error(f"Simulation {sim_data.get('status', '')}: {sim_data}")
                return AlphaSimulationResult.fail(expr=expression, sim_id=sim_id, message="Simulation failed,response: " + sim_data)
            logger.info(f"Simulation complete, checking alpha {alpha_id}")
            alpha_data = self.load_alpha_result(alpha_id)
            return AlphaSimulationResult.success(expr=expression, sim_id=sim_id, sim_data=sim_data, alpha_data=alpha_data)
        except Exception as e:
            logger.error(f"Error monitoring simulation {sim_id}: {str(e)}")
            return AlphaSimulationResult.fail(expr=expression, sim_id=sim_id, message=str(e))

    def load_alpha_result(self, alpha_id: str):
        assert alpha_id != ''
        excep = None
        for i in range(3):
            try:
                alpha_response = self._session_manager.get(f"https://api.worldquantbrain.com/alphas/{alpha_id}")
                if alpha_response.status_code == 200:
                    return alpha_response.json()
                else:
                    time.sleep(5)
            except Exception as e:
                excep = e
                time.sleep(5)
        if excep is None:
            raise Exception('Load alpha response fail.')
        else:
            raise excep


    def wait_for_simulation_complete(self, sim_id: str):
        sim_progress_url = f"https://api.worldquantbrain.com/simulations/{sim_id}"
        try:
            while True:
                sim_progres_resp = self._session_manager.get(sim_progress_url)
                sim_data = sim_progres_resp.json()
                if "progress" in sim_data:
                    progress = sim_data.get("progress", 0)
                    logger.info(f"Simulation progress: {progress:.2%}")
                retry_after_sec = float(sim_progres_resp.headers.get("Retry-After", 0))
                # simulation done
                if retry_after_sec == 0:
                    break
                # sleep at least 10 secs
                time.sleep(max(retry_after_sec, 10.0))
            return sim_progres_resp.json()
        except Exception as e:
            logger.error(f"Error testing alpha: {str(e)}")
            raise e

    def execute(self, expression: str, settings: SimulatorSettings = None) -> AlphaSimulationResult:
        if settings is None:
            settings = SimulatorSettings()
        submit_res = self.submit_simulation(expression=expression, settings=settings)
        if submit_res.success:
            return self.monitor_simulation(sim_id=submit_res.sim_id, expression=expression)
        else:
            return AlphaSimulationResult.fail(expr=expression, sim_id = '', message=submit_res.message)

    def execute_multi(self, expression_list: List[str], settings: SimulatorSettings = None, rate_limiter: RateLimiter = None) -> List[AlphaSimulationResult]:
        if settings is None:
            settings = SimulatorSettings()
        if rate_limiter is None:
            rate_limiter = RateLimiter()
        logging.info(f"Starting batch simulation of {len(expression_list)} alphas")
        alpha_results: List[AlphaSimulationResult] = []
        with ThreadPoolExecutor(max_workers = 4) as executor:
            futures = []
            for expression in expression_list:
                rate_limiter.wait_for_slot()
                future = executor.submit(self.execute, expression=expression, settings=settings)
                futures.append(future)
            # process results and set up monitoring
            for future in futures:
                try:
                    alpha_result = future.result()
                    if alpha_result:
                        alpha_results.append(alpha_result)
                except Exception as e:
                    logging.error(f"Error in simulating alpha: {str(e)}")
            return alpha_results



if __name__ == '__main__':
    s = Simulator(session_manager=SessionManager())
    res = s.execute(expression="rank(open - (ts_mean(vwap, 10))) * (-1 * abs(rank(close - vwap)))")
    print(res)
    # periods = [10, 20, 33, 66]
    # text_exprs = []
    # for period in periods:
    #     expr = f"rank(open - (ts_mean(vwap, {period}))) * (-1 * abs(rank(close - vwap)))"
    #     print(expr)
    #     text_exprs.append(expr)
    # for period in periods:
    #     expr = f"rank(open - (ts_mean(close, {period}))) * (-1 * abs(rank(open - vwap)))"
    #     text_exprs.append(expr)
    # res = s.execute_multi(expression_list=text_exprs)
    # print(json.dumps(res))





import logging

from brainminer.brain import SessionManager, WorldQuantBrain, SimulatorSettings, PerformanceFilter, \
    AlphaSimulationResult
from typing import List

class AlphaMiner:

    def __init__(self, session_manager: SessionManager):
        self._brain = WorldQuantBrain(session_manager)

    def mine_alphas(self, alpha_list: List[str], settings: SimulatorSettings, perf_filter: PerformanceFilter) -> List[AlphaSimulationResult]:
        logging.info(f"Starting alpha mining, size: {len(alpha_list)}")
        alpha_results = self._brain.execute_multi_simulation(expression_list=alpha_list, settings=settings)
        logging.info(f"Alpha mining completed, size: {len(alpha_results)}")
        match_results = []
        for alpha_result in alpha_results:
            if alpha_result.success and alpha_result.alpha_data is not None and alpha_result.alpha_data.is_match(perf_filter):
                match_results.append(alpha_result)
        return match_results


import json
import pandas as pd
import itertools
from typing import List, Dict

def df_to_markdown(df: pd.DataFrame, columns=None) -> str:
    """
    将 pandas.DataFrame 导出为 Markdown 表格字符串，可以指定导出的列名。

    参数:
    df (pd.DataFrame): 要导出的 DataFrame。
    columns (list, optional): 要导出的列名。如果为 None，则导出所有列。

    返回:
    str: Markdown 表格字符串。
    """
    if columns is not None:
        df = df[columns]

    # 构建表头
    header = '| ' + ' | '.join(df.columns) + ' |'
    separator = '| ' + ' | '.join(['---'] * len(df.columns)) + ' |'

    # 构建表格内容
    rows = []
    for index, row in df.iterrows():
        rows.append('| ' + ' | '.join(map(str, row)) + ' |')

    # 合并所有部分
    markdown_table = '\n'.join([header, separator] + rows)
    return markdown_table

def df_to_json(df: pd.DataFrame, columns=None) -> str:
    if columns is not None:
        df = df[columns]
    # 将 DataFrame 转换为 JSON 数组
    json_array = df.to_dict(orient='records')
    return json.dumps(json_array, ensure_ascii=False)

def is_empty_metrics(metrics):
    if metrics is None:
        return True
    if metrics['sharpe'] is None and metrics['fitness'] is None:
        return True
    return False

def build_alpha(template: str, params: Dict) -> List[str]:
    """
    所有可能的表达式
    :param template: 表达式模板
    :param params: 参数说明，key为对应要替换的变量，value为列表
    :return:
    """
    keys = list(params.keys())
    values = list(params.values())
    # 生成笛卡尔积
    combinations = list(itertools.product(*values))
    # 替换模板中的变量
    result = []
    for combination in combinations:
        result_dict = dict(zip(keys, combination))
        result.append(template.format(**result_dict))
    return result